==> constrain-0.1.2.tar.gz: lib/src/runtime_constraint_resolver.dart <==
// Copyright (c) 2014, The Constrain project authors.
// Please see the AUTHORS file for details.
// All rights reserved. Use of this source code is governed by
// a BSD 2-Clause License that can be found in the LICENSE file.

library constraint.metadata.resolver.runtime;

import '../constraint.dart';
import 'package:constrain/metadata.dart';
import 'dart:mirrors';
import 'package:option/option.dart';
import 'preconditions.dart';
import 'util.dart';

final TypeMirror _iterableType = reflectType(Iterable);
final TypeMirror _mapType = reflectType(Map);

class RuntimeTypeDescriptorResolver implements TypeDescriptorResolver {
  final Map _cache = <Type, Option<TypeDescriptor>>{};

  // TODO: need some other way to key functions when we move to code
  // generation
  final Map _functionCache = <MethodMirror, Option<FunctionDescriptor>>{};

  // TODO: Does Dart have weak references?

  @override
  Option<TypeDescriptor> resolveFor(Type clazz) {
    if (clazz == null) {
      return const None();
    }

    return _resolveForClass(reflectClass(clazz));
  }


  @override
  Option<FunctionDescriptor> resolveForFunction(Function function) {
    if (function == null) {
      return const None();
    }

    return _resolveForFunctionOrMethod((reflect(function) as ClosureMirror).function);
  }

  Option<TypeDescriptor> _resolveForClass(ClassMirror cm) {
    if (cm == null || !cm.hasReflectedType) {
      return const None();
    }

    final cachedDescriptor = _cache[cm.reflectedType];

    if (cachedDescriptor != null) {
      return cachedDescriptor;
    }

    // TODO: is there any risk of concurrent modification. i.e. different futures
    // triggering this?

    /*
     * To cope with type loops we put a proxy in to start with. If we hit a
     * type loop then we will end up actually linking up the proxy to a member.
     * So when we have the real type, we first wire the real descriptor into
     * the proxy before we replace it in the cache. The vast majority of
     * members will be wired into a real descriptor rather than a proxy
     */
    final proxy = new _TypeDescriptorProxy();
    _cache[cm.reflectedType] = new Some(proxy);

    final newDescriptor = _resolveForClassUncached(cm);
    if (newDescriptor.nonEmpty()) {
      proxy.realTypeDescriptor = newDescriptor.get();
    }

    _cache[cm.reflectedType] = newDescriptor;

    return newDescriptor;
  }

  Option<TypeDescriptor> _resolveForClassUncached(ClassMirror cm) {

    if (cm.isTopLevel && cm.owner.simpleName == #dart.core) {
      return const None();
    }

    final Set<MemberDescriptor> memberDescriptors = _resolveMembers(cm);

    Option<TypeDescriptor> superclassDescriptor = _resolveSuperclass(cm);

    Iterable<TypeDescriptor> superinterfaceDescs = _resolveSuperinterfaces(cm);

    final Set<ConstraintDescriptor> classConstraints =
        _resolveConstraints(cm);

    final Set<FunctionDescriptor> methodDescriptors = _resolveMethods(cm);


    final hasConstraints =
        memberDescriptors.isNotEmpty
        || methodDescriptors.isNotEmpty
        || classConstraints.isNotEmpty
        || superclassDescriptor.nonEmpty()
        || superinterfaceDescs.isNotEmpty;

    return hasConstraints ?
        new Some(new TypeDescriptor(cm.reflectedType, classConstraints,
            memberDescriptors, methodDescriptors,
            superclassDescriptor, superinterfaceDescs))
        : const None();
  }

  Option<TypeDescriptor> _resolveSuperclass(ClassMirror cm) {
    return _resolveForClass(cm.superclass);
  }

  Iterable<TypeDescriptor> _resolveSuperinterfaces(ClassMirror cm) {
    return cm.superinterfaces.map((i) =>
        optionAsIterable(_resolveForClass(i)))
        .expand((e) => e);
  }

  Set<MemberDescriptor> _resolveMembers(ClassMirror cm) {
    final memberDescriptors = new Set<MemberDescriptor>();

    cm.declarations.forEach((k, v) {
      if ((v is VariableMirror && !v.isStatic) ||
          (v is MethodMirror && !v.isStatic && (v.isGetter /* || v.isSetter */))) {

        final TypeMirror rawMemberType = v is VariableMirror ? v.type :
          v is MethodMirror ? v.returnType : throw new StateError('oops');

        var memberDescriptor = _resolveMember(v, rawMemberType, k);

        if (memberDescriptor.nonEmpty()) {
          memberDescriptors.add(memberDescriptor.get());
        }
      }
    });

    return memberDescriptors;
  }

  Set<FunctionDescriptor> _resolveMethods(ClassMirror cm) {
    final methodDescriptors = new Set<FunctionDescriptor>();

    // TODO: constructors etc
    cm.declarations.forEach((k, v) {
      if (v is MethodMirror && !v.isStatic && !v.isGetter && !v.isSetter
          && !v.isConstructor) {

        var methodDescriptor = _resolveForFunction(v);

        if (methodDescriptor.nonEmpty()) {
          methodDescriptors.add(methodDescriptor.get());
        }
      }
    });

    return methodDescriptors;
  }



  Option<MemberDescriptor> _resolveMember(DeclarationMirror dm,
      TypeMirror rawMemberType, Symbol name) {
    final Set<ConstraintDescriptor> constraintDescriptors =
        _resolveConstraints(dm);

    final isCollection = _isCollection(rawMemberType);
    final isGeneric = _isGeneric(rawMemberType);

    if (isCollection && !isGeneric) {
      return const None(); // nothing to go on here
    }

    // TODO: not sure what to do about more than one type arg.
    // Maybe this should be treated same as no generics?
    final memberType = !isCollection ? rawMemberType :
      rawMemberType.typeArguments.first;

    final Option<TypeDescriptor> memberTypeDescriptor =
        _resolveForType(memberType);

    return (constraintDescriptors.isNotEmpty ||
        memberTypeDescriptor.nonEmpty()) ?
            new Some(new MemberDescriptor(name, constraintDescriptors,
                  memberTypeDescriptor,
                  isMultivalued: isCollection, isGeneric: isGeneric))
        : const None();
  }

  Option<TypeDescriptor> _resolveForType(TypeMirror tm) {
    return tm is ClassMirror ? _resolveForClass(tm) : const None();
  }

  Set<ConstraintDescriptor> _resolveConstraints(DeclarationMirror dm) {
    return dm.metadata
        .where((InstanceMirror im) => im.reflectee is Constraint)
        .map((im) => im.reflectee)
        .map((c) => new ConstraintDescriptor(c))
        .toSet();
  }

  bool _isCollection(TypeMirror tm) =>
      tm is ClassMirror && tm.isSubtypeOf(_iterableType) ||
        tm.isSubtypeOf(_mapType);

  bool _isGeneric(TypeMirror tm) =>  tm is ClassMirror &&
    tm.typeArguments.isNotEmpty;


  Option<FunctionDescriptor> _resolveForFunctionOrMethod(MethodMirror mm) {
    if (mm.owner is ClassMirror) {
      return _resolveForClass(mm.owner).flatMap((td) =>
          new Option(td.methodDescriptorsMap[mm.simpleName]));
    }
    return _resolveForFunction(mm);
  }


  Option<FunctionDescriptor> _resolveForFunction(MethodMirror mm) {
    if (mm == null) {
      return const None();
    }

    final cachedDescriptor = _functionCache[mm];

    if (cachedDescriptor != null) {
      return cachedDescriptor;
    }

    final newDescriptor = _resolveForFunctionUncached(mm);

    _functionCache[mm] = newDescriptor;

    return newDescriptor;
  }





  Option<FunctionDescriptor> _resolveForFunctionUncached(MethodMirror mm) {
    if (mm.isTopLevel && mm.owner.simpleName == #dart.core) {
      return const None();
    }

    final positionalMemberDescriptors =
        _resolveParameters(mm, false);

    final List<Option<MemberDescriptor>> namedMemberDescriptors =
        _resolveParameters(mm, true);


    final returnDescriptor = _resolveReturn(mm);

    final hasConstraints =
        positionalMemberDescriptors.any((o) => o.nonEmpty())
        || namedMemberDescriptors.any((o) => o.nonEmpty())
        || returnDescriptor.nonEmpty()
//        || superclassFunctionDescriptor.nonEmpty()
//        || superinterfaceFunctionDescs.isNotEmpty
        ;

    final namedParameterDescriptorsMap = <Symbol, MemberDescriptor>{};

    namedMemberDescriptors.forEach((mdOpt) {
      if (mdOpt.nonEmpty()) {
        final md = mdOpt.get();
        namedParameterDescriptorsMap[md.name] = md;
      }
    });


    return hasConstraints ?
        new Some(new FunctionDescriptor(mm.simpleName,
            positionalMemberDescriptors,
            namedParameterDescriptorsMap, returnDescriptor
//            superclassFunctionDescriptor, superinterfaceFunctionDescs
            ))
        : const None();

  }

  List<Option<MemberDescriptor>> _resolveParameters(MethodMirror mm, bool named) {
    return mm.parameters.where((pm) => pm.isNamed == named)
        .map((pm) {
      return _resolveMember(pm, pm.type, pm.simpleName);
    }).toList(growable: false);
  }

  // TODO: dodgy name - ugly
  Option<MemberDescriptor> _resolveReturn(MethodMirror mm) =>
      _resolveMember(mm, mm.returnType, #___return);


}

class _TypeDescriptorProxy implements TypeDescriptor {
  TypeDescriptor _realTypeDescriptor;

  TypeDescriptor get realTypeDescriptor {
    // to prevent it being called before it's initialised
    ensure(_realTypeDescriptor, isNotNull);
    return _realTypeDescriptor;
  }

  set realTypeDescriptor(TypeDescriptor d) {
    _realTypeDescriptor = d;
  }

  @override
  Set<ConstraintDescriptor> get constraintDescriptors =>
      realTypeDescriptor.constraintDescriptors;

  @override
  Iterable<MemberDescriptor> get memberDescriptors =>
      realTypeDescriptor.memberDescriptors;

  @override
  Map<Symbol, MemberDescriptor> get memberDescriptorsMap =>
    realTypeDescriptor.memberDescriptorsMap;

  @override
  Type get type => realTypeDescriptor.type;

  @override
  Map<Symbol, FunctionDescriptor> get methodDescriptorsMap =>
      realTypeDescriptor.methodDescriptorsMap;
}

// TODO: need to:
// - support inheritence of types
// - difference between simple types and classes
//   - i.e. classes have super types, simple types won't have their own type constraints

==> constrain-0.1.4.tar.gz: lib/src/runtime_constraint_resolver.dart <==
// Copyright (c) 2014, The Constrain project authors.
// Please see the AUTHORS file for details.
// All rights reserved. Use of this source code is governed by
// a BSD 2-Clause License that can be found in the LICENSE file.

library constraint.metadata.resolver.runtime;

import '../constraint.dart';
import 'package:constrain/metadata.dart';
import 'dart:mirrors';
import 'package:option/option.dart';
import 'preconditions.dart';
import 'util.dart';
import 'dart:async';

final TypeMirror _iterableType = reflectType(Iterable);
final TypeMirror _mapType = reflectType(Map);
final TypeMirror _futureType = reflectType(Future);


class RuntimeTypeDescriptorResolver implements TypeDescriptorResolver {
  final Map _cache = <Type, Option<TypeDescriptor>>{};

  // TODO: need some other way to key functions when we move to code
  // generation
  final Map _functionCache = <MethodMirror, Option<FunctionDescriptor>>{};

  // TODO: Does Dart have weak references?

  @override
  Option<TypeDescriptor> resolveFor(Type clazz) {
    if (clazz == null) {
      return const None();
    }

    return _resolveForClass(reflectClass(clazz));
  }


  @override
  Option<FunctionDescriptor> resolveForFunction(Function function) {
    if (function == null) {
      return const None();
    }

    return _resolveForFunctionOrMethod((reflect(function) as ClosureMirror).function);
  }

  Option<TypeDescriptor> _resolveForClass(ClassMirror cm) {
    if (cm == null || !cm.hasReflectedType) {
      return const None();
    }

    final cachedDescriptor = _cache[cm.reflectedType];

    if (cachedDescriptor != null) {
      return cachedDescriptor;
    }

    // TODO: is there any risk of concurrent modification. i.e. different futures
    // triggering this?

    /*
     * To cope with type loops we put a proxy in to start with. If we hit a
     * type loop then we will end up actually linking up the proxy to a member.
     * So when we have the real type, we first wire the real descriptor into
     * the proxy before we replace it in the cache. The vast majority of
     * members will be wired into a real descriptor rather than a proxy
     */
    final proxy = new _TypeDescriptorProxy();
    _cache[cm.reflectedType] = new Some(proxy);

    final newDescriptor = _resolveForClassUncached(cm);
    if (newDescriptor.nonEmpty()) {
      proxy.realTypeDescriptor = newDescriptor.get();
    }

    _cache[cm.reflectedType] = newDescriptor;

    return newDescriptor;
  }

  Option<TypeDescriptor> _resolveForClassUncached(ClassMirror cm) {

    if (cm.isTopLevel && cm.owner.simpleName == #dart.core) {
      return const None();
    }

    final Set<MemberDescriptor> memberDescriptors = _resolveMembers(cm);

    Option<TypeDescriptor> superclassDescriptor = _resolveSuperclass(cm);

    Iterable<TypeDescriptor> superinterfaceDescs = _resolveSuperinterfaces(cm);

    final Set<ConstraintDescriptor> classConstraints =
        _resolveConstraints(cm);

    final Set<FunctionDescriptor> methodDescriptors = _resolveMethods(cm);


    final hasConstraints =
        memberDescriptors.isNotEmpty
        || methodDescriptors.isNotEmpty
        || classConstraints.isNotEmpty
        || superclassDescriptor.nonEmpty()
        || superinterfaceDescs.isNotEmpty;

    return hasConstraints ?
        new Some(new TypeDescriptor(cm.reflectedType, classConstraints,
            memberDescriptors, methodDescriptors,
            superclassDescriptor, superinterfaceDescs))
        : const None();
  }

  Option<TypeDescriptor> _resolveSuperclass(ClassMirror cm) {
    return _resolveForClass(cm.superclass);
  }

  Iterable<TypeDescriptor> _resolveSuperinterfaces(ClassMirror cm) {
    return cm.superinterfaces.map((i) =>
        optionAsIterable(_resolveForClass(i)))
        .expand((e) => e);
  }

  Set<MemberDescriptor> _resolveMembers(ClassMirror cm) {
    final memberDescriptors = new Set<MemberDescriptor>();

    cm.declarations.forEach((k, v) {
      if ((v is VariableMirror && !v.isStatic) ||
          (v is MethodMirror && !v.isStatic && (v.isGetter /* || v.isSetter */))) {

        final TypeMirror rawMemberType = v is VariableMirror ? v.type :
          v is MethodMirror ? v.returnType : throw new StateError('oops');

        var memberDescriptor = _resolveMember(v, rawMemberType, k);

        if (memberDescriptor.nonEmpty()) {
          memberDescriptors.add(memberDescriptor.get());
        }
      }
    });

    return memberDescriptors;
  }

  Set<FunctionDescriptor> _resolveMethods(ClassMirror cm) {
    final methodDescriptors = new Set<FunctionDescriptor>();

    // TODO: constructors etc
    cm.declarations.forEach((k, v) {
      if (v is MethodMirror && !v.isStatic && !v.isGetter && !v.isSetter
          && !v.isConstructor) {

        var methodDescriptor = _resolveForFunction(v);

        if (methodDescriptor.nonEmpty()) {
          methodDescriptors.add(methodDescriptor.get());
        }
      }
    });

    return methodDescriptors;
  }



  Option<MemberDescriptor> _resolveMember(DeclarationMirror dm,
      TypeMirror rawMemberType, Symbol name) {
    final Set<ConstraintDescriptor> constraintDescriptors =
        _resolveConstraints(dm);

    final isCollection = _isCollection(rawMemberType);
    final isFuture = _isFuture(rawMemberType);
    final isGeneric = _isGeneric(rawMemberType);

    if ((isFuture || isCollection) && !isGeneric) {
      return const None(); // nothing to go on here
    }

    // TODO: not sure what to do about more than one type arg.
    // Maybe this should be treated same as no generics?
    final memberType = !(isCollection || isFuture) ? rawMemberType :
      rawMemberType.typeArguments.first;

    final Option<TypeDescriptor> memberTypeDescriptor =
        _resolveForType(memberType);

    return (constraintDescriptors.isNotEmpty ||
        memberTypeDescriptor.nonEmpty()) ?
            new Some(new MemberDescriptor(name, constraintDescriptors,
                  memberTypeDescriptor,
                  isMultivalued: isCollection, isGeneric: isGeneric,
                  isFuture: isFuture))
        : const None();
  }

  Option<TypeDescriptor> _resolveForType(TypeMirror tm) {
    return tm is ClassMirror ? _resolveForClass(tm) : const None();
  }

  Set<ConstraintDescriptor> _resolveConstraints(DeclarationMirror dm) {
    return dm.metadata
        .where((InstanceMirror im) => im.reflectee is Constraint)
        .map((im) => im.reflectee)
        .map((c) => new ConstraintDescriptor(c))
        .toSet();
  }

  bool _isCollection(TypeMirror tm) =>
      tm is ClassMirror && tm.isSubtypeOf(_iterableType) ||
        tm.isSubtypeOf(_mapType);

  bool _isFuture(TypeMirror tm) =>
      tm is ClassMirror && tm.isSubtypeOf(_futureType);

  bool _isGeneric(TypeMirror tm) =>  tm is ClassMirror &&
    tm.typeArguments.isNotEmpty;


  Option<FunctionDescriptor> _resolveForFunctionOrMethod(MethodMirror mm) {
    if (mm.owner is ClassMirror) {
      return _resolveForClass(mm.owner).flatMap((td) =>
          new Option(td.methodDescriptorsMap[mm.simpleName]));
    }
    return _resolveForFunction(mm);
  }


  Option<FunctionDescriptor> _resolveForFunction(MethodMirror mm) {
    if (mm == null) {
      return const None();
    }

    final cachedDescriptor = _functionCache[mm];

    if (cachedDescriptor != null) {
      return cachedDescriptor;
    }

    final newDescriptor = _resolveForFunctionUncached(mm);

    _functionCache[mm] = newDescriptor;

    return newDescriptor;
  }





  Option<FunctionDescriptor> _resolveForFunctionUncached(MethodMirror mm) {
    if (mm.isTopLevel && mm.owner.simpleName == #dart.core) {
      return const None();
    }

    final positionalMemberDescriptors =
        _resolveParameters(mm, false);

    final List<Option<MemberDescriptor>> namedMemberDescriptors =
        _resolveParameters(mm, true);


    final returnDescriptor = _resolveReturn(mm);

    final hasConstraints =
        positionalMemberDescriptors.any((o) => o.nonEmpty())
        || namedMemberDescriptors.any((o) => o.nonEmpty())
        || returnDescriptor.nonEmpty()
//        || superclassFunctionDescriptor.nonEmpty()
//        || superinterfaceFunctionDescs.isNotEmpty
        ;

    final namedParameterDescriptorsMap = <Symbol, MemberDescriptor>{};

    namedMemberDescriptors.forEach((mdOpt) {
      if (mdOpt.nonEmpty()) {
        final md = mdOpt.get();
        namedParameterDescriptorsMap[md.name] = md;
      }
    });


    return hasConstraints ?
        new Some(new FunctionDescriptor(mm.simpleName,
            positionalMemberDescriptors,
            namedParameterDescriptorsMap, returnDescriptor
//            superclassFunctionDescriptor, superinterfaceFunctionDescs
            ))
        : const None();

  }

  List<Option<MemberDescriptor>> _resolveParameters(MethodMirror mm, bool named) {
    return mm.parameters.where((pm) => pm.isNamed == named)
        .map((pm) {
      return _resolveMember(pm, pm.type, pm.simpleName);
    }).toList(growable: false);
  }

  // TODO: dodgy name - ugly
  Option<MemberDescriptor> _resolveReturn(MethodMirror mm) =>
      _resolveMember(mm, mm.returnType, #___return);


}

class _TypeDescriptorProxy implements TypeDescriptor {
  TypeDescriptor _realTypeDescriptor;

  TypeDescriptor get realTypeDescriptor {
    // to prevent it being called before it's initialised
    ensure(_realTypeDescriptor, isNotNull);
    return _realTypeDescriptor;
  }

  set realTypeDescriptor(TypeDescriptor d) {
    _realTypeDescriptor = d;
  }

  @override
  Set<ConstraintDescriptor> get constraintDescriptors =>
      realTypeDescriptor.constraintDescriptors;

  @override
  Iterable<MemberDescriptor> get memberDescriptors =>
      realTypeDescriptor.memberDescriptors;

  @override
  Map<Symbol, MemberDescriptor> get memberDescriptorsMap =>
    realTypeDescriptor.memberDescriptorsMap;

  @override
  Type get type => realTypeDescriptor.type;

  @override
  Map<Symbol, FunctionDescriptor> get methodDescriptorsMap =>
      realTypeDescriptor.methodDescriptorsMap;
}

// TODO: need to:
// - support inheritence of types
// - difference between simple types and classes
//   - i.e. classes have super types, simple types won't have their own type constraints
